import 'dart:async';
import 'package:flutter/services.dart';
import 'package:interactive_webview_null_safety/interactive_webview.dart';
import 'package:uuid/uuid.dart';
import 'message_broker.dart';
import 'dart:convert';
import 'dto.dart';
import 'player/tivio_player_wrapper.dart';
import 'streams.dart';
import 'logger.dart';

class CallFunctionErrorPayload {
  String? messageId;
  final String type;

  CallFunctionErrorPayload({
    this.messageId,
    type,
  }) : type = type ?? 'data';
}

class Bridge {
  static final Bridge _instance = Bridge._internal();
  final _uuid = Uuid();
  final _webView = InteractiveWebView();
  final _jsonEncoder = JsonEncoder();
  final _dataMessageBroker = DataMessageBroker();
  final _callbackMessageBroker = CallbackMessageBroker();
  final _logger = Logger();
  var initConfiguration = {};
  StreamSubscription? webviewMessageStream;
  StreamSubscription? webviewStateStream;

  factory Bridge() {
    return _instance;
  }

  Bridge._internal();

  loadBridgeJs() {
    PlatformAssetBundle loader = PlatformAssetBundle();
    return loader.loadString('packages/tivio_sdk/assets/bridge.build7.js');
  }

  /// TD: remove after resolving TIV-434
  Future<void> initWithoutInitialization(String appId, String language, bool development) async {
    try {
      _logger.debug('Tivio bridge: initialization started');

      initConfiguration['appId'] = appId;
      initConfiguration['language'] = language;
      initConfiguration['development'] = development;

      if (webviewMessageStream != null) {
        webviewMessageStream!.cancel();
      }

      if (webviewStateStream != null) {
        webviewStateStream!.cancel();
      }

      var _bridgeReady = Completer();
      _logger.debug('Tivio bridge: loading of js build started');
      String bundle = await loadBridgeJs();
      _logger.debug('Tivio bridge: loading of js build completed. Bundle size: ${bundle.length}');

      webviewMessageStream = _webView.didReceiveMessage.listen((event) {
        switch (event.data['type']) {
          case 'data':
            _dataMessageBroker.publish(
              event.data['messageId'],
              event.data['data'],
            );
            break;
          case 'callback':
            _callbackMessageBroker.publish(
              event.data['messageId'],
              event.data['data'],
            );
            break;
        }
      }, onError: (error) => _logger.error(error));

      webviewStateStream = _webView.stateChanged.listen((event) {
        if (event.type == WebViewState.didFinish) {
          _logger.debug('Tivio bridge: WebView did finish');
          _bridgeReady.complete();
        }
      }, onError: (error) => _logger.error(error));

      await _webView.loadHTML('''
<!DOCTYPE html>
<html>
  <head>
    <meta charset="UTF-8">
    <script>
      $bundle
    </script>
  </head>
  <body>
  </body>
</html>
''', baseUrl: 'https://tiv.io');

      _logger.debug('Tivio bridge: initialization ended');
      return _bridgeReady.future;
    } catch (exception) {
      _logger.error(exception);
    }
  }

  /// TD: remove after resolving TIV-434
  Future<void> initWithoutBridge(String appId, String language, bool development) async {
    try {
      _logger.debug('Tivio bridge: initialization started');

      initConfiguration['appId'] = appId;
      initConfiguration['language'] = language;
      initConfiguration['development'] = development;

      if (webviewMessageStream != null) {
        webviewMessageStream!.cancel();
      }

      if (webviewStateStream != null) {
        webviewStateStream!.cancel();
      }

      var _bridgeReady = Completer();

      webviewMessageStream = _webView.didReceiveMessage.listen((event) {
        print('event: ${event.data['type']}, ${event.data['messageId']}, ${event.data['data']}');
        switch (event.data['type']) {
          case 'data':
            _dataMessageBroker.publish(
              event.data['messageId'],
              event.data['data'],
            );
            break;
          case 'callback':
            _callbackMessageBroker.publish(
              event.data['messageId'],
              event.data['data'],
            );
            break;
        }
      }, onError: (error) => _logger.error(error));

      webviewStateStream = _webView.stateChanged.listen((event) {
        if (event.type == WebViewState.didFinish) {
          _logger.debug('Tivio bridge: WebView did finish');
        }
      }, onError: (error) => _logger.error(error));

      _logger.debug('Tivio bridge: initialization ended');
      _bridgeReady.complete();
      return _bridgeReady.future;
    } catch (exception) {
      _logger.error(exception);
    }
  }

  Future<void> init(String appId, String language, bool development) async {
    try {
      _logger.debug('Tivio bridge: initialization started');

      initConfiguration['appId'] = appId;
      initConfiguration['language'] = language;
      initConfiguration['development'] = development;

      if (webviewMessageStream != null) {
        webviewMessageStream!.cancel();
      }

      if (webviewStateStream != null) {
        webviewStateStream!.cancel();
      }

      var _bridgeReady = Completer();
      _logger.debug('Tivio bridge: loading of js build started');
      String bundle = await loadBridgeJs();
      _logger.debug('Tivio bridge: loading of js build completed. Bundle size: ${bundle.length}');

      webviewMessageStream = _webView.didReceiveMessage.listen((event) {
        switch (event.data['type']) {
          case 'data':
            _dataMessageBroker.publish(
              event.data['messageId'],
              event.data['data'],
            );
            break;
          case 'callback':
            _callbackMessageBroker.publish(
              event.data['messageId'],
              event.data['data'],
            );
            break;
        }
      }, onError: (error) => _logger.error(error));

      webviewStateStream = _webView.stateChanged.listen((event) {
        if (event.type == WebViewState.didFinish) {
          _logger.debug('Tivio bridge: WebView did finish');
          callFunction(
            'init',
            [appId, language, development],
            (payload) {
              _logger.debug('Tivio bridge: initialization completed');
              _bridgeReady.complete();
            },
          );
        }
      }, onError: (error) => _logger.error(error));

      await _webView.loadHTML('''
<!DOCTYPE html>
<html>
  <head>
    <meta charset="UTF-8">
    <script>
      $bundle
    </script>
  </head>
  <body>
  </body>
</html>
''', baseUrl: 'https://tiv.io');

      _logger.debug('Tivio bridge: initialization ended');
      return _bridgeReady.future;
    } catch (exception) {
      _logger.error(exception);
    }
  }

  getPaginatedData<T>(
    String functionName,
    List<dynamic> functionArgs,
    T Function(dynamic record) createDto,
  ) {
    return callFunction<TivioDataPagination<T>>(
      functionName,
      functionArgs,
      (payload) {
        return TivioDataPagination<T>(
          (payload['records'] != null) ? payload['records'].map<T>((record) => createDto(record)).toList() : [],
          payload['hasNextPage'] ?? false,
        );
      },
    );
  }

  Future<T?> callFunction<T>(String functionName, List<dynamic> functionArgs, T Function(dynamic payload)? onComplete, { CallFunctionErrorPayload? errorPayload }) {
    Completer done = Completer<T>();

    String messageId = _uuid.v4();
    functionArgs.add(messageId);
    String serializedArgs = _jsonEncoder.convert(functionArgs);

    _webView.evalJavascript('''
try {
  window.TivioBridge.$functionName($serializedArgs);
} catch (error) {
  console.error('Error in evaluating javascript');
  console.error(error);

  const nativeCommunicator = typeof webkit !== 'undefined' ? webkit.messageHandlers.native : window.native;
  nativeCommunicator.postMessage(JSON.stringify({
    messageId: '${(errorPayload != null && errorPayload.messageId != null) ? errorPayload.messageId : messageId}',
    data: {
      error: {
        code: 'TIVIO_IS_NOT_INITIALIZED',
        message: error.message,
        stack: error.stack,
        fileName: error.fileName,
        lineNumber: error.lineNumber,
      },
    },
    type: '${(errorPayload != null) ? errorPayload.type : 'data'}',
  }));
}
''');


    _dataMessageBroker.listen(messageId, (payload) => done.complete((onComplete != null) ? onComplete(payload) : null));

    return done.future as Future<T?>;
  }

  handleNewEventInStream <T>(dynamic payload, TivioDataStreamController<T> controller, void Function() callFunctionWithArguments, T Function (dynamic payload) parseData) async {
    if (payload['error'] != null) {
      _logger.error(payload['error']);

      if (payload['error']['code'] == 'TIVIO_IS_NOT_INITIALIZED') {
        _logger.warning('Re-initializing Tivio');

        await init(initConfiguration['appId'], initConfiguration['language'], initConfiguration['development']);

        callFunctionWithArguments();
      } else {
        controller.addErrorSafely(ArgumentError(payload['error']['message']));
      }
    } else {
      controller.addSafely(parseData(payload));
    }
  }

  setUser(String userId, dynamic payload, {dynamic additionalUserData}) {
    return callFunction('setUser', [userId, payload, additionalUserData], (payload) => null);
  }

  setLanguage(String language) {
    return callFunction('setLanguage', [language], (payload) => null);
  }

  /// Player

  getPlayer(String id) {
    return callFunction<TivioPlayerWrapper>(
      'getPlayer',
      [id],
      (payload) => TivioPlayerWrapper(id, this),
    );
  }

  playerRegister(String id) {
    return callFunction('playerRegister', [id], (payload) => null);
  }

  playerOnSourceChanged(String id, data) {
    return callFunction(
      'playerOnSourceChanged',
      [id, _jsonEncoder.convert(data)],
      (payload) => null,
    );
  }

  playerPlay(String id) {
    return callFunction('playerPlay', [id], (payload) => {});
  }

  playerPause(String id) {
    return callFunction('playerPause', [id], (payload) => {});
  }

  playerSeekTo(String id, num timeInMilliseconds) {
    return callFunction('playerSeekTo', [id, timeInMilliseconds], (payload) => {});
  }

  playerSeekByStep(String id, num timeInMilliseconds) {
    return callFunction('playerSeekByStep', [id, timeInMilliseconds], (payload) => {});
  }

  playerOnTimeChanged(String id, num timestampMs) {
    return callFunction(
      'playerOnTimeChanged',
      [id, timestampMs],
      (payload) => null,
    );
  }

  playerOnStateChanged(String id, String state) {
    return callFunction('playerOnStateChanged', [id, state], (payload) => null);
  }

  onPlaybackEnded(String id) {
    return callFunction('playerOnPlaybackEnded', [id], (payload) => null);
  }

  playerSkipAd(String playerId) {
    return callFunction('playerSkipAd', [playerId], (payload) => null);
  }

  onPlayerPlay(String id, Function callback) {
    _callbackMessageBroker.listen('player-$id-play', callback);
  }

  onPlayerPause(String id, Function callback) {
    _callbackMessageBroker.listen('player-$id-pause', callback);
  }

  onPlayerChangeSource(String id, Function callback) {
    _callbackMessageBroker.listen('player-$id-set-source', callback);
  }

  onPlayerSeekTo(String id, Function callback) {
    _callbackMessageBroker.listen('player-$id-seek-to', callback);
  }

  /// API 1.0.0

  /// Widget
  TivioDataStream<TivioDataWidget>? subscribeToWidget({
    required String widgetId,
  }) {
    var controller = TivioDataStreamController<TivioDataWidget>();
    var callFunctionWithArguments = () => callFunction('subscribeToWidget', [widgetId], (payload) => null, errorPayload: CallFunctionErrorPayload(messageId: 'subscribe-to-widget:$widgetId', type: 'callback'));

    _callbackMessageBroker.listen('subscribe-to-widget:$widgetId', (payload) {
      handleNewEventInStream(payload, controller, callFunctionWithArguments, (payload) => TivioDataWidget.fromJson(payload));
    });

    callFunctionWithArguments();

    return controller.tivioDataStream;
  }

  /// Channel

  TivioDataStream<TivioDataPagination<TivioDataChannel>>? subscribeToChannelsInWidget({
    required String widgetId,
    required int count,
  }) {
    var controller = TivioDataStreamController<TivioDataPagination<TivioDataChannel>>();
    var callFunctionWithArguments = () => callFunction('subscribeToChannelsInWidget', [widgetId, count], (payload) => null, errorPayload: CallFunctionErrorPayload(messageId: 'subscribe-to-channels-in-widget:$widgetId', type: 'callback'));

    _callbackMessageBroker.listen('subscribe-to-channels-in-widget:$widgetId', (payload) {
      handleNewEventInStream(payload, controller, callFunctionWithArguments, (payload) => TivioDataChannel.fromJsonPaginated(payload));
    });

    callFunctionWithArguments();

    return controller.tivioDataStream;
  }

  fetchNextChannelsInWidget({
    required String widgetId,
    required int count,
  }) {
    callFunction('fetchNextChannelsInWidget', [widgetId, count], (payload) => null, errorPayload: CallFunctionErrorPayload(messageId: 'subscribe-to-channels-in-widget:$widgetId', type: 'callback'));
  }

  TivioDataStream<TivioDataChannel>? subscribeToChannel({required String channelId}) {
    var controller = TivioDataStreamController<TivioDataChannel>();
    var callFunctionWithArguments = () => callFunction('subscribeToChannel', [channelId], (payload) => null, errorPayload: CallFunctionErrorPayload(messageId: 'subscribe-to-channel:$channelId', type: 'callback'));

    _callbackMessageBroker.listen('subscribe-to-channel:$channelId', (payload) {
      handleNewEventInStream(payload, controller, callFunctionWithArguments, (payload) => TivioDataChannel.fromJson(payload));
    });

    callFunctionWithArguments();

    return controller.tivioDataStream;
  }

  /// Section

  TivioDataStream<TivioDataPagination<TivioDataSection>>? subscribeToSectionsInChannel({
    required String channelId,
    required int count,
  }) {
    var controller = TivioDataStreamController<TivioDataPagination<TivioDataSection>>();
    var callFunctionWithArguments = () => callFunction('subscribeToSectionsInChannel', [channelId, count], (payload) => null, errorPayload: CallFunctionErrorPayload(messageId: 'subscribe-to-sections-in-channel:$channelId', type: 'callback'));

    _callbackMessageBroker.listen('subscribe-to-sections-in-channel:$channelId', (payload) {
      handleNewEventInStream(payload, controller, callFunctionWithArguments, (payload) => TivioDataSection.fromJsonPaginated(payload));
    });

    callFunctionWithArguments();

    return controller.tivioDataStream;
  }

  fetchNextSectionsInChannel({
    required String channelId,
    required int count,
  }) {
    callFunction('fetchNextSectionsInChannel', [channelId, count], (payload) => null, errorPayload: CallFunctionErrorPayload(messageId: 'subscribe-to-sections-in-channel:$channelId', type: 'callback'));
  }

  TivioDataStream<TivioDataSection>? subscribeToSection({required String sectionId}) {
    var controller = TivioDataStreamController<TivioDataSection>();
    var callFunctionWithArguments = () => callFunction('subscribeToSection', [sectionId], (payload) => null, errorPayload: CallFunctionErrorPayload(messageId: 'subscribe-to-section:$sectionId', type: 'callback'));

    _callbackMessageBroker.listen('subscribe-to-section:$sectionId', (payload) {
      handleNewEventInStream(payload, controller, callFunctionWithArguments, (payload) => TivioDataSection.fromJson(payload));
    });

    callFunctionWithArguments();

    return controller.tivioDataStream;
  }

  /// Videos

  TivioDataStream<TivioDataPagination<TivioDataVideo>>? subscribeToVideosInSection({
    required String sectionId,
    required int count,
  }) {
    var controller = TivioDataStreamController<TivioDataPagination<TivioDataVideo>>();
    var callFunctionWithArguments = () => callFunction('subscribeToVideosInSection', [sectionId, count], (payload) => null, errorPayload: CallFunctionErrorPayload(messageId: 'subscribe-to-videos-in-section:$sectionId', type: 'callback'));

    _callbackMessageBroker.listen('subscribe-to-videos-in-section:$sectionId', (payload) {
      handleNewEventInStream(payload, controller, callFunctionWithArguments, (payload) => TivioDataVideo.fromJsonPaginated(payload));
    });

    callFunctionWithArguments();

    return controller.tivioDataStream;
  }

  fetchNextVideosInSection({
    required String sectionId,
    required int count,
  }) {
    callFunction('fetchNextVideosInSection', [sectionId, count], (payload) => null, errorPayload: CallFunctionErrorPayload(messageId: 'subscribe-to-videos-in-section:$sectionId', type: 'callback'));
  }

  TivioDataStream<TivioDataVideo>? subscribeToVideo({required String videoId}) {
    var controller = TivioDataStreamController<TivioDataVideo>();
    var callFunctionWithArguments = () => callFunction('subscribeToVideo', [videoId], (payload) => null, errorPayload: CallFunctionErrorPayload(messageId: 'subscribe-to-video:$videoId', type: 'callback'));

    _callbackMessageBroker.listen('subscribe-to-video:$videoId', (payload) {
      handleNewEventInStream(payload, controller, callFunctionWithArguments, (payload) => TivioDataVideo.fromJson(payload));
    });

    callFunctionWithArguments();

    return controller.tivioDataStream;
  }

  /// Recent

  TivioDataStream<TivioDataPagination<TivioDataVideo>>? subscribeToRecentVideosInWidget({
    required String widgetId,
    required int count,
  }) {
    var controller = TivioDataStreamController<TivioDataPagination<TivioDataVideo>>();

    controller.addSafely(TivioDataPagination([], false));

    return controller.tivioDataStream;
  }

  TivioDataStream<TivioDataPagination<TivioDataVideo>>? subscribeToRecentVideosInChannel({
    required String channelId,
    required int count,
  }) {
    var controller = TivioDataStreamController<TivioDataPagination<TivioDataVideo>>();

    controller.addSafely(TivioDataPagination([], false));

    return controller.tivioDataStream;
  }

  /// Search

  searchChannels(String query, int limit) {
    return getPaginatedData<TivioDataChannel>(
      'searchChannels',
      [query, limit],
      (record) => TivioDataChannel.fromJson(record),
    );
  }

  searchVideos(String query, int limit) {
    return getPaginatedData<TivioDataVideo>(
      'searchVideos',
      [query, limit],
      (record) => TivioDataVideo.fromJson(record),
    );
  }

  /// Screen

  TivioDataStream<TivioDataScreen>? subscribeToScreen({
    required String screenId,
  }) {
    var controller = TivioDataStreamController<TivioDataScreen>();
    var callFunctionWithArguments = () => callFunction('subscribeToScreen', [screenId], (payload) => null, errorPayload: CallFunctionErrorPayload(messageId: 'subscribe-to-screen:$screenId', type: 'callback'));

    _callbackMessageBroker.listen('subscribe-to-screen:$screenId', (payload) {
      handleNewEventInStream(payload, controller, callFunctionWithArguments, (payload) => TivioDataScreen.fromJson(payload));
    });

    callFunctionWithArguments();

    return controller.tivioDataStream;
  }

  TivioDataStream<TivioDataPagination<TivioDataRowItem>>? subscribeToItemsInRow({
    required String rowId,
    required int count,
  }) {
    var controller = TivioDataStreamController<TivioDataPagination<TivioDataRowItem>>();
    var callFunctionWithArguments = () => callFunction('subscribeToItemsInRow', [rowId, count], (payload) => null, errorPayload: CallFunctionErrorPayload(messageId: 'subscribe-to-items-in-row:$rowId', type: 'callback'));

    _callbackMessageBroker.listen('subscribe-to-items-in-row:$rowId', (payload) {
      handleNewEventInStream(payload, controller, callFunctionWithArguments, (payload) => TivioDataRowItem.fromJsonPaginated(payload));
    });

    callFunctionWithArguments();

    return controller.tivioDataStream;
  }

  fetchNextItemsInRow({
    required String rowId,
    required int count,
  }) {
    callFunction('fetchNextItemsInRow', [rowId, count], (payload) => null, errorPayload: CallFunctionErrorPayload(messageId: 'subscribe-to-items-in-row:$rowId', type: 'callback'));
  }
}
