class TivioDataWidget {
  final String id;
  final String name;

  TivioDataWidget(this.id, this.name);

  TivioDataWidget.fromJson(Map json)
      : id = json['id'],
        name = json['name'];
}

class _ChannelImages {
  final String? logo;
  final String? cover;

  _ChannelImages(this.logo, this.cover);
}

class TivioDataChannel {
  final String id;
  final String name;
  final _ChannelImages images;

  TivioDataChannel(this.id, this.name, this.images);

  TivioDataChannel.fromJson(Map json)
      : id = json['id'],
        name = json['name'],
        images = _ChannelImages(json['headerLogo'], json['cover']);

  static TivioDataPagination<TivioDataChannel> fromJsonPaginated(Map json) {
    var records = json['records'] ?? [];
    var channels = records
        .where((record) => (record['id'] != null && record['name'] != null))
        .map<TivioDataChannel>((record) => TivioDataChannel.fromJson(record))
        .toList();

    return TivioDataPagination(channels, json['hasNextPage'] ?? false);
  }
}

class TivioDataSection {
  final String id;
  final String name;

  TivioDataSection(this.id, this.name);

  TivioDataSection.fromJson(Map json)
      : id = json['id'],
        name = json['name'];

  static TivioDataPagination<TivioDataSection> fromJsonPaginated(Map data) {
    var records = data['records'] ?? [];
    var sections = records
        .where((record) => (record['id'] != null && record['name'] != null))
        .map<TivioDataSection>((record) => TivioDataSection.fromJson(record))
        .toList();

    return TivioDataPagination(sections, data['hasNextPage'] ?? false);
  }
}

class _VideoImages {
  final String? cover;

  _VideoImages(this.cover);
}

class _Monetization {
  final num? price;
  final String? type;

  _Monetization(this.price, this.type);
}

class TivioDataPlaylistItem {
  final String name;
  final String url;
  final Duration duration;

  TivioDataPlaylistItem(this.name, this.url, this.duration);
}

class TivioDataVideo {
  final String id;
  final String name;
  final Duration duration;
  final DateTime? created;
  final String? url;
  final String? adMonetizationId;
  final _VideoImages images;
  final bool isPlaylist;
  final List<TivioDataPlaylistItem> playlistItems;
  final _Monetization? monetization;

  TivioDataVideo(
    this.id,
    this.name,
    this.duration,
    this.created,
    this.url,
    this.adMonetizationId,
    this.images,
    this.isPlaylist,
    this.playlistItems,
    this.monetization,
  );

  TivioDataVideo.fromJson(Map json)
      : id = json['id'],
        name = json['name'],
        duration = Duration(milliseconds: json['duration'] ?? 0),
        created = (json['created'] != null) ? DateTime.parse(json['created']) : null,
        url = json['url'],
        adMonetizationId = json['adMonetizationId'],
        images = _VideoImages(json['cover']),
        isPlaylist = json['isPlaylist'] ?? false,
        playlistItems = [],
        monetization = (json['monetization'] != null)
            ? _Monetization(json['monetization']['name'], json['monetization']['price'])
            : null;

  static TivioDataPagination<TivioDataVideo> fromJsonPaginated(Map data) {
    var records = data['records'] ?? [];
    var videos = records
        .where((record) => (record['id'] != null && record['name'] != null))
        .map<TivioDataVideo>((record) => TivioDataVideo.fromJson(record))
        .toList();

    return TivioDataPagination(videos, data['hasNextPage'] ?? false);
  }
}

class TivioAssetScale {
  final String? background;

  TivioAssetScale({
    this.background,
  });
}

class TivioAssets {
  final TivioAssetScale small;
  final TivioAssetScale medium;
  final TivioAssetScale large;

  TivioAssets({
    required this.small,
    required this.medium,
    required this.large,
  });
}

Map<String, TivioAssets> parseAssets (Map? assets) {
  var rawAssets = assets ?? Map();
  var parsedAssets = Map<String, TivioAssets>();

  rawAssets.forEach((name, asset) {
    parsedAssets[name] = TivioAssets(
      small: TivioAssetScale(
        background: asset['@1']['background'],
      ),
      medium: TivioAssetScale(
        background: asset['@2']['background'],
      ),
      large: TivioAssetScale(
        background: asset['@3']['background'],
      ),
    );
  });

  return parsedAssets;
}

class TivioDataScreen {
  final String id;
  final String name;
  final List<TivioDataRow> rows;
  final Map<String, TivioAssets> assets;
  final String? description;

  TivioDataScreen({
    required this.id,
    required this.name,
    required this.rows ,
    required this.assets,
    this.description,
  });

  static TivioDataScreen fromJson(Map json) {
    var rawRows = json['rows'] ?? [];
    var parsedRows = rawRows
        .where((record) => (record['id'] != null && record['name'] != null))
        .map<TivioDataRow>((record) => TivioDataRow.fromJson(record))
        .toList();

    return TivioDataScreen(
        id: json['id'],
        name: json['name'],
        rows: parsedRows,
        assets: parseAssets(json['assets']),
        description: json['description'],
    );
  }
}

class TivioDataRow {
  final String id;
  final String name;
  final String itemComponent;
  final String rowComponent;
  final Map<String, TivioAssets> assets;
  final String? description;

  TivioDataRow({
    required this.id,
    required this.name,
    required this.itemComponent,
    required this.rowComponent,
    required this.assets,
    this.description,
  });

  TivioDataRow.fromJson(Map json)
      : id = json['id'],
        name = json['name'],
        itemComponent = json['itemComponent'],
        rowComponent = json['rowComponent'],
        assets = parseAssets(json['assets']),
        description = json['description'];
}

enum TivioRowItemType {
  Video,
  Tag,
}

class TivioDataRowItem {
  final TivioRowItemType type;
  final String name;
  final Map<String, TivioAssets> assets;
  final String? description;
  final TivioDataVideo? video;
  final TivioDataTag? tag;

  TivioDataRowItem({
    required this.type,
    required this.name,
    required this.assets,
    this.description,
    this.video,
    this.tag,
  });

  TivioDataRowItem.fromJson(Map json)
      : name = json['name'],
        type = TivioRowItemType.Video,
        assets = parseAssets(json['assets']),
        description = json['description'],
        video = TivioDataVideo.fromJson(json),
        tag = null;

  static TivioDataPagination<TivioDataRowItem> fromJsonPaginated(Map data) {
    var records = data['records'] ?? [];
    var items = records
        .where((record) => (record['name'] != null))
        .map<TivioDataRowItem>((record) => TivioDataRowItem.fromJson(record))
        .toList();

    return TivioDataPagination(items, false);
  }
}

class TivioDataTag {

}

class TivioDataPagination<T> {
  final List<T> records;
  final bool hasNextPage;

  TivioDataPagination(this.records, this.hasNextPage);
}
